
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
extern "C" void abort(void);
extern "C" void exit(int status);
#endif

#include 	<stdio.h>

#ifdef HAVE_MALLOC_H
#include 	<malloc.h>
#endif

#include        "compilerApi.h"
#include        "strategy.h"

#include	"basic_fun.h"
#include	"global_var.h"
#include        "trans2.h"

//char	*get_file_out(char *);
char *newOutputFileName(char *inputFileName);
extern void intrin_exp(SgStatement *,SgStatement *,SgExpression *);
extern int expression_extract(SgStatement *,SgStatement *);
extern int expression_deal(SgStatement *,SgStatement *,SgExpression *);
extern void extract_intrinsic_function(SgStatement *,SgStatement *);
extern int array_to_forall(SgStatement *,SgStatement *);
extern int element_to_array(SgStatement *,SgStatement *);
extern int where_to_forall(SgStatement *,SgStatement *);

//added by slx
extern int normalize_forallconstruct_header(SgStatement *, SgStatement *);
extern int forallconstruct_to_multi_forallconstruct(SgStatement *, SgStatement *);
extern int forallconstruct_to_forall(SgStatement *, SgStatement *);
extern void trans_ele_intrinsic(SgStatement *);

SgVariableSymb	*temp_index[8];
SgValueExp 	*base_step;//default step is 1
SgValueExp	*base_lbound;
SgValueExp	*default_lsub;//Fortran default low bound is 1
SgStatement     *current;//current statement to be transformed
int		index_g;//digit for forall index variable. 
int		index_var;//digit for temporary variable.

PassType passtype;

enum Flag { off, on };
Flag in_main_program = off;
Flag in_procedure = off;
Flag in_function = off;
Flag in_module = off;

int do_level = 0;

int element_intrinsic_call = STRATEGY_ELEMENT_INTRINSIC_CALL;

int silentFlag = 0;

/* This function is used to cut the surfix, so there is a record as different
pass performed */
char *newOutputFileName(char *inputFileName)
{
  char  *out;
  int   nameLength = strlen(inputFileName);
  
  out=(char *)malloc((unsigned)(nameLength+5));

  while (inputFileName[nameLength-1]!='.') {
    nameLength --;
  }

  for ( int i = 0; i<nameLength; i++) {
    out[i]=inputFileName[i];
  }

  out[nameLength-1]='\0';

  out=strcat(out,"_2.dep");

  return(out);
}

void analyse(SgStatement * s)
{
  if (s->numberOfAttributes(MUST_DELETE) > 0) {
    s->deleteStmt();
    return;
  }

  switch(s->variant()) {

  case GLOBAL:
    {
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);
      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;

      break;
    };

  case PROG_HEDR:
    {
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i; 

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      // do nothing?;
      in_main_program=on;

	// do some initialization 
      init_temp_index(s);

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;

      in_main_program=off;
	
      break;

    };

  case PROC_HEDR:
    {
      SgProcHedrStmt * prochedr = (SgProcHedrStmt *)s;
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      in_procedure=on;
	// do some initialization 
      init_temp_index(s);

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;
      
      in_procedure = off;

      break;
    };

  case FUNC_HEDR:
    {
      SgProcHedrStmt * prochedr = (SgProcHedrStmt *)s;
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      in_procedure=on;
	// do some initialization 
      init_temp_index(s);

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;
      
      in_procedure = off;

      break;
    };

  case MODULE_STMT:
    {
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i; 

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      // do nothing?;
      in_module=on;

	// do some initialization 
      init_temp_index(s);

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;

      in_module=off;
	
      break;

    };

  case FOR_NODE:
    {
      do_level++;

      int org_numberOfChildren=s->numberOfChildrenList1();

      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;
      
      do_level--;

      break;
    }

  case LOGIF_NODE:
  case ELSEIF_NODE:
  case LOGIF_STMT:
  case SWITCH_NODE:
  case CASE_NODE:
  case DO_WHILE_NODE:
    {
      if (passtype == ASSIGN_PASS)
      {
	SgStatement *hedr_stmt = s->getScopeForDeclare();
	ASSERT(hedr_stmt);
	// extract array expression argument as an assign statement 
	SgExpression *con = s->expr(0);
	expression_deal(hedr_stmt,s,con);
      }

      int org_numberOfChildren1=s->numberOfChildrenList1();

      SgStatement * * children1 = new SgStatement * [org_numberOfChildren1];

      int org_numberOfChildren2=s->numberOfChildrenList2();

      SgStatement * * children2 = new SgStatement * [org_numberOfChildren2];

      int i;

      for (i=0; i<org_numberOfChildren1; i++)
	children1[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren1; i++)
	analyse(children1[i]);
      
      delete [] children1;

      for (i=0; i<org_numberOfChildren2; i++)
	children2[i]=s->childList2(i);			// zheng.gb

      for (i=0; i<org_numberOfChildren2; i++)
	analyse(children2[i]);
      
      delete [] children2;

      break;

    }

  case TASK_STMT:
  case BLOCK_RESIDENT_STMT:
  case BLOCK_ON_STMT: {
      int org_numberOfChildren=s->numberOfChildrenList1();

      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;
      
      break;
    }

  case ASSIGN_STAT:
    {
	if (passtype == PRE_ASSIGN_PASS ) {
		if ( element_intrinsic_call == EXTRACT_ELEMENT_INTRINSIC_CALL) {
			SgStatement *hedr_stmt = s->getScopeForDeclare();
			ASSERT(hedr_stmt);
			extract_intrinsic_function(hedr_stmt,s);
		}
        }

	// change array element to array section like: a(1) => a(1:1)
	// this will leave the assign of array element to trans0
	if (passtype == ASSIGN_PASS)
	{
	SgStatement *hedr_stmt = s->getScopeForDeclare();
	ASSERT(hedr_stmt);
	// extract array expression argument as an assign statement 
	expression_extract(hedr_stmt,s);
	// change array element of lhs to array section like: a(i) => a(i:i)
/* now I deal this directly in assign.cc of trans0
	element_to_array(hedr_stmt,s);
*/
	//change array assign statement to foral statement
        array_to_forall(hedr_stmt,s);
	}
        if (passtype == ELEMENTARY_INTRINSIC_PASS)
        {
	SgStatement *hedr_stmt = s->getScopeForDeclare();
	ASSERT(hedr_stmt);
        trans_ele_intrinsic(s); 
        }
        break;
    }

  case PROC_STAT:
    {
	if (passtype == ASSIGN_PASS)
	{
	SgStatement *hedr_stmt = s->getScopeForDeclare();
	ASSERT(hedr_stmt);
	// extract array expression argument as an assign statement 
	intrin_exp(hedr_stmt,s,s->expr(0));
	}
	break;
    }
  case WHERE_NODE:
    {
	if (passtype == WHERE_STMT_PASS)
	{
	SgStatement *hedr_stmt = s->getScopeForDeclare();
	ASSERT(hedr_stmt);
        where_to_forall(hedr_stmt,s);
	}
        break;
    }

  case WHERE_BLOCK_STMT:
    {
extern int whereconstruct_to_where(SgStatement *,SgStatement *);
	if (passtype == WHERE_CONSTRUCT_PASS)
	{
	SgStatement *hedr_stmt = s->getScopeForDeclare();
	ASSERT(hedr_stmt);
        whereconstruct_to_where(hedr_stmt,s);
	}
        break;
    }

  case FORALL_STMT:
    {       
      int org_numberOfChildren1=s->numberOfChildrenList1();

      SgStatement * * children1 = new SgStatement * [org_numberOfChildren1];

      int i;

      for (i=0; i<org_numberOfChildren1; i++)
	children1[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren1; i++)
	analyse(children1[i]);
      
	//  if is FORALL_STMT , means we have translate the assign statement to FORALL
      if (s->childList1(0)->variant() == FORALL_STMT  || s->numberOfChildrenList1() > 2 )
		s->thebif->variant = FORALL_CONSTRUCT_STMT;

      delete [] children1;

      break;
    }

  case FORALL_CONSTRUCT_STMT:
    {
       if (passtype == FORALL_CONSTRUCT_PASS_1)
        {
	 SgStatement *hedr_stmt = s->getScopeForDeclare();
	 ASSERT(hedr_stmt);
         normalize_forallconstruct_header(hedr_stmt,s);
        } 
	else if (passtype == FORALL_CONSTRUCT_PASS_2)
	{
 	 SgStatement *hedr_stmt = s->getScopeForDeclare();
       	 ASSERT(hedr_stmt);
         forallconstruct_to_multi_forallconstruct(hedr_stmt,s);
        } 
        else if(passtype == FORALL_CONSTRUCT_PASS_3)
	{
 	 SgStatement *hedr_stmt = s->getScopeForDeclare();
	 ASSERT(hedr_stmt);
         forallconstruct_to_forall(hedr_stmt,s);
        } 
         else {
      int org_numberOfChildren1=s->numberOfChildrenList1();

      SgStatement * * children1 = new SgStatement * [org_numberOfChildren1];

      int i;

      for (i=0; i<org_numberOfChildren1; i++)
	children1[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren1; i++)
	analyse(children1[i]);
      
      delete [] children1;
      }
	break;
    }

  case SIMPLE_RESIDENT_STMT: {
      if (passtype == OTHER_PASS) {
        SgStatement *next = s->nextInChildList();
        ASSERT(next);
        SgStatement *next1 = next->copyBlockPtr();
	SgStatement *block = new SgBlockResidentStmt(*s->expr(0)->copyPtr(), *next1);
        s->insertStmtBefore(*block, *s->controlParent());
	s->deleteStmt();
	next->addAttribute(MUST_DELETE, 0, 0);
      }
    break;
    }

  case SIMPLE_ON_STMT: {
      if (passtype == OTHER_PASS) {
        SgStatement *next = s->nextInChildList();
        ASSERT(next);
        SgStatement *next1 = next->copyPtr();
	SgExpression *op1 = s->expr(0)?s->expr(0)->copyPtr():(SgExpression *)NULL;
	SgExpression *op2 = s->expr(1)?s->expr(1)->copyPtr():(SgExpression *)NULL;
	SgExpression *op3 = s->expr(2)?s->expr(2)->copyPtr():(SgExpression *)NULL;
	SgStatement *block = new SgBlockOnStmt(op1, op2, op3, *next1);
        s->insertStmtBefore(*block, *s->controlParent());
	s->deleteStmt();
	next->addAttribute(MUST_DELETE, 0, 0);
      }
    break;
    }

  case PRINT_STAT:
  case WRITE_STAT:
//    analyse_datatransfer(s);
    break;

  case CONTROL_END:
  case PARAM_DECL:
  case ENDDO_STMT:
  case  FORMAT_STAT:
   // statement need do nothing;
    break;

  default:
//    printf("Not touched statement in analyse on line %d\n", s->lineNumber());
    break;
  };

}

void transform_1(SgStatement * s) 
{
  switch(s->variant()) {
  case GLOBAL:
    {
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);
      for (i=0; i<org_numberOfChildren; i++)
	transform_1(children[i]);
      
      delete [] children;

      break;
    };

  case PROG_HEDR:
    {
      in_main_program = on;
      add_new_dec(s);
      in_main_program = off;

      break;
    };

  case PROC_HEDR:
    {
      in_procedure = on;
      add_new_dec(s);      
      in_procedure = off;

      break;
    };

  case FUNC_HEDR:
    {
      in_procedure = on;
      add_new_dec(s);      
      in_procedure = off;

      break;
    };

  case MODULE_STMT:
    {
      in_module = on;
      add_new_dec(s);
      in_module = off;

      break;
    };

  };
}

static void usage()
{
        printf("Usage: trans2 [-h|--help] [-z] HPF_file.dep \n");
        printf("options:\n");
        printf("   -h|--help   : help message. \n");
        printf("   -z          : suppress all compiling message.\n");
}

void parse_string(char *s)
{
        if (strncmp(s,"help",4) == 0) {
                usage();
                exit(0);
        }
        else 
                printf("warning : Unknown option \"%s\" ignored.\n",s);
}

int main(int argc, char **argv)
{

//  	SgProject 	project("test", argc, argv);
	SgProject       *project;
 	SgFile	*file=NULL;

  	SgStatement	*hedr_stmt,*next_stmt,*end_stmt;
	int	i=0,numoffunc,numoffiles;
	char	*file_in,*file_out;

	PassType   pt[10];
	int no=0;

	argv++;
        argc--;
	if (argc == 0) { usage(); exit(1); }
        while ((argc > 0) && ((*argv)[0] == '-') )
        {
                i=1;
                while (argv[0][i] != '\0') {
                        switch (argv[0][i]) {
                        case 'a'  :
				pt[no++]=WHERE_CONSTRUCT_PASS;
				break;
                        case 'b'  :
				pt[no++]=WHERE_STMT_PASS;
				break;
                        case 'c'  :
				pt[no++]=PRE_ASSIGN_PASS;
				pt[no++]=ASSIGN_PASS;
				break;
                        case 'd' :
                                pt[no++]=ELEMENTARY_INTRINSIC_PASS;
                                break;
                        case 'e'  :
                                pt[no++]=FORALL_CONSTRUCT_PASS_1;
                                break; 
                        case 'f'  :
                                pt[no++]=FORALL_CONSTRUCT_PASS_2;
                                break;
                        case 'g'  :
                                pt[no++]=FORALL_CONSTRUCT_PASS_3;
                                break;
                        case 'h'  :
                                pt[no++]=OTHER_PASS;
                                break;
                        case 'z'  :
                                silentFlag = 1;
                                break;
			case '-':
                                if (i==1) {
                                        parse_string(argv[0]+i+1);
                                        i=strlen(argv[0])-1;   /* quit loop */
                                }
                                break;
			default:
        			printf("warning 422: Unknown option \"%c\" ignored.\n",argv[0][i]);
        			exit(1);
			}
			i++;
		}
		argc--;
                argv++;
	}
	if ( no == 0 )
	{
		pt[no++]=WHERE_CONSTRUCT_PASS;
		pt[no++]=WHERE_STMT_PASS;
		pt[no++]=PRE_ASSIGN_PASS;
		pt[no++]=ASSIGN_PASS;
                
                //added by slx
                pt[no++]=ELEMENTARY_INTRINSIC_PASS;
                pt[no++]=FORALL_CONSTRUCT_PASS_1;
                pt[no++]=FORALL_CONSTRUCT_PASS_2;
                pt[no++]=FORALL_CONSTRUCT_PASS_3;

                pt[no++]=OTHER_PASS;
	}

	argc++; argv--;
	project = new SgProject("test", argc, argv);
	SgStatement *s;

	numoffiles=project->numberOfFiles();
	for(i=0;i<numoffiles;i++)
	{
		base_lbound=new SgValueExp(0);
		base_step=new SgValueExp(1);
		default_lsub=new SgValueExp(1);//fortran default low bound is 1
		file=&(project->file(i));	
		file_in=file->filept->filename;
  		if (file->languageType()!=ForSrc)
		{	printf("%s is  not fortran program.\n",file_in); 
			continue;
		}

		char comment[100];
		for (int j=0; j<no ; j++)
    		{
			passtype=pt[j];
			if (silentFlag == 0)
			{
			switch (passtype) {
        		case WHERE_CONSTRUCT_PASS :
                		sprintf(comment,"%s","Where construct ");
                		break;
        		case WHERE_STMT_PASS :
                		sprintf(comment,"%s","Where statement ");
                		break;
        		case PRE_ASSIGN_PASS :
                		sprintf(comment,"%s","Pre-Assign statement");
                		break;
        		case ASSIGN_PASS :
                		sprintf(comment,"%s","Assign statement");
                		break;
                        case ELEMENTARY_INTRINSIC_PASS:
                                sprintf(comment,"%s","Elementary intrinsic");
                                break;
                         // added by slx
                        case FORALL_CONSTRUCT_PASS_1:
                                sprintf(comment,"%s","Forall construct 1");
                                break;
                        case FORALL_CONSTRUCT_PASS_2:
                                sprintf(comment,"%s","Forall construct 2");
                                break; 
                        case FORALL_CONSTRUCT_PASS_3:
                                sprintf(comment,"%s","Forall construct 3");
                                break; 
                        case OTHER_PASS:
                                sprintf(comment,"%s","Other transformations");
                                break; 
			}
			printf("***********  PASS %s **********\n",comment);
			}
			s = file->firstStatement();
			analyse(s);
		}

		transform_1(s);

		file_out=newOutputFileName(file_in);
		file->saveDepFile(file_out);
		//save the transformation result (AST) to file result*.dep

	}
  	return 0;
}

